/**
   Integrator class is a collection of static routines for integration
   purposes. It contains the folowing methods: <br/>

   1. trapezoidal(Function f, double a, double b, tolerance) <br/>
   2. simpsonRule(Function f, double a, double b, tolerance) <br/>
   3. midpoint(Function f, double a, double b, tolerance) <br/>
   4. Romberg(Function f, double a, double b, tolerance) <br/>
   5. gaussLegendre(Function f, double a, double b, tolerance) <br/>

   @see <a href="http://en.wikipedia.org/wiki/Romberg%27s_method">Wikipedia</a>
   <br/> and Grant Palmer.
   <em>Technical Java™: Developing Scientific and Engineering Applications</em>
   Publisher: Prentice Hall PTR, 2003.
 */

public class Integrator {

    // Error code if there is no convergence
    static double ERRCODE = 1234567.;

    /**
       Trapezoidal rule of integration up to a given tolerance
       @param function Function type object implementing Function
       @param start start of integration interval
       @param end end of integration interval
       @param tolerance given accuracy
    */
    public static double trapezoidal(Function function,
                                     double start, double end,
                                     double tolerance) {
        double dx = end - start;
        double value, oldValue, sum, scale;
        int maxIter = 20;
        // Compute an initial value for the integral using
        // the two-point trapezoidal rule.
        value = 0.5*dx*( function.getValue(start) +
                         function.getValue(end) );
        // Subdivide the integration range by adding more
        // points, first 1 point, then two more, then four
        // more and so on. Recompute the integral value
        // each time. If the solution converges, return the
        // value.
        for(int i=0; i<maxIter; ++i) {
            sum = 0.0;
            scale = Math.pow(0.5,i+1);
            for(int j=0;j<Math.pow(2,i); ++j) {
                sum += function.getValue( start + scale*(2*j + 1.0)*dx);
            }
            oldValue = value; value = 0.5*value + scale*dx*sum;
            if ( Math.abs(value-oldValue) <= tolerance*Math.abs(oldValue) ) {
                return value;
            }
        }
        // Solution did not converge
        return ERRCODE;
    }


    /**
       Simpson rule of integration up to a given tolerance
       @param function Function type object implementing Function
       @param start start of integration interval
       @param end end of integration interval
       @param tolerance given accuracy
    */
    public static double simpsonsRule(Function function,
                                      double start, double end,
                                      double tolerance) {
        double dx = end - start;
        double value, oldValue, sum, scale;
        double valueSimpson, oldValueSimpson;
        int maxIter = 20;
        // Set initial values for the trapezoidal
        // method (value) and Simpson's rule (valueSimpson)
        value = 0.5*dx*( function.getValue(start) + function.getValue(end) );
        valueSimpson = 1.0e+6;
        // Subdivide the integration range by adding more
        // points, first one point, then two more, then
        // four more, and so on. Recompute the extended
        // trapezoidal value and the Simpson's rule value.
        // If the Simpson's rule value converges, return
        // the value.
        for(int i=0; i<maxIter; ++i) {
            sum = 0.0; scale = Math.pow(0.5,i+1);
            for(int j=0; j<Math.pow(2,i); ++j) {
                sum += function.getValue( start + scale*(2*j + 1.0)*dx);
            }
            oldValue = value;
            value = 0.5*value + scale*dx*sum;
            oldValueSimpson = valueSimpson;
            valueSimpson = (4.0*value - oldValue)/3.0;
            if ( Math.abs(valueSimpson-oldValueSimpson) <=
                 tolerance*Math.abs(oldValueSimpson) ) {
                return valueSimpson;
            }
        }  // Solution did not converge
        return ERRCODE;
    }


    /** Romberg rule of integration by extrapolation method
        @see: http://en.wikipedia.org/wiki/Romberg%27s_method
        @param function Function type object implementing Function
        @param a integration interval starting value
        @param b integration interval end
    */
    public static double Romberg( Function function, double a, double b,
                                  double tolerance ) {
        int maxIter = 25;
        int minIter = 3;
        double r[][] = new double[maxIter][maxIter];

        r[0][0] = 0.5*(b-a)*(function.getValue(a)+function.getValue(b));
        int twon = 1;
        for (int n=1; n<maxIter; n++) {
                twon *= 2;
                int kmax = twon/2;
                double hn = (b-a)/twon;
                r[n][0] = 0.5*r[n-1][0];
                for (int k=1; k<=kmax; k++)
                    r[n][0] += hn * function.getValue(a+(2*k-1)*hn);
                for (int m=1; m<=n; m++){
                    double fourm = Math.pow(4,m);
                    r[n][m] = (fourm*r[n][m-1] - r[n-1][m-1])/(fourm-1);
                    //System.out.print(r[n][m]);
                }
                //System.out.println("");
                double err = Math.abs(r[n][n] - r[n][n-1]);
                if ((n>=minIter) && (err < tolerance))
                    return r[n][n];
            }
        return ERRCODE;
    }


    /**
       Open midpoint rule of integration up to a given tolerance
       @param function, Function type object implementing Function
       @param start start of integration interval
       @param end end of integration interval
       @param tolerance given accuracy
    */
    public static double midpoint(Function function,
                                  double start, double end, double tolerance) {
        double dx = end - start;
        double value, oldValue, sum, temp;
        double scale = 1.0/3.0;
        int numPoints, numPairs;
        int maxIter = 20;
        // Start with a two-point midpoint evaluation.
        value = 0.5*dx*( function.getValue(start + 0.25*dx) +
                         function.getValue(end - 0.25*dx) );
        // Refine the solution by adding points to the
        // integration domain. Each iteration adds three
        // times as many points as the previous iteration.
        // When the solution converges, return the result.
        for(int i=0; i<maxIter; ++i) {
            numPoints = 4*(int)Math.pow(3,i);
            numPairs = (numPoints - 2)/2;
            temp = 1.0/(3.0*numPoints);
            // Add the two endpoint values to the sum
            sum = function.getValue(start + temp*dx) +
                function.getValue(end - temp*dx);
            // Add in each pair to the sum
            for(int j=0; j<numPairs; ++j) {
                sum = sum + function.getValue(start + (6*j + 5)*temp*dx) +
                    function.getValue(start + (6*j + 7)*temp*dx);
            }
            oldValue = value;
            value = value/3.0 + 0.5*Math.pow(scale,i+1)*dx*sum;
            if ( Math.abs(value-oldValue) <= tolerance*Math.abs(oldValue) ) {
                return value;
            }
        }
        // Solution did not converge
        return ERRCODE;
    }


    /**
       The weights and abscissa values are normalized
       for an integration range between 0 and 1. The
       values are symmetric about x=0.5 so you only
       have to compute half of them.
       @param x[] quadrature nodes, array
       @param w[] quadrature weights, array
       @param numPoints number of Gauss integration points
    */
    public static void computeGaussWeights( double x[], double w[],
                                            int numPoints)
    {
        int i,j,m;
        double z, zOld, root, p3, p2, p1;
        double EPS = 1.0e-10;

        m = (numPoints + 1)/2;

        // Compute the Legendre polynomial, p1, at point
        // z using Newton's method.
        for(i=0; i<m; ++i) {
            z = Math.cos(Math.PI*(i+0.75)/(numPoints+0.5));
            do { p1 = 1.0; p2 = 0.0;
                for(j=0; j<numPoints; ++j) {
                    p3 = p2;
                    p2 = p1;
                    p1 = ((2.0*(j+1) - 1.0)*z*p2 - j*p3)/ (j+1.0);
                }
                root = numPoints*(z*p1 - p2)/(z*z - 1.0);
                zOld = z; z = zOld - p1/root;
            } while ( Math.abs(z-zOld) > EPS);
            // Compute the weights and abscissas. The values
            // are symmetric about the point z=0.5
            x[i] = 0.5*(1.0 - z);
            x[numPoints-1-i] = 0.5*(1.0 + z);
            w[i] = 1.0/((1.0- z*z)*root*root);
            w[numPoints-1-i] = w[i];
        }
    }


    /**
       Gauss-Legendre integration up to a given tolerance
       @param function Function type object implementing Function
       @param start start of integration interval
       @param end end of integration interval
       @param tolerance given accuracy
    */
    public static double gaussLegendre(Function function,
                                       double start, double end,
                                       double tolerance)
    {
        double dx = end - start;
        double value = 1.0e+10;
        double oldValue;
        int numPoints = 10;
        int maxPoints = 100000;
        double x[] = new double[maxPoints];
        double w[] = new double[maxPoints];

        // Compute the integral value adding points
        // until the solution converges or maxPoints is reached.
        while(true) {
            // Calculate the Gauss-Legendre weights and abscissas
            computeGaussWeights(x, w, numPoints);
            // Compute the integral value by summing up the
            // weights times the function value at each abscissa
            oldValue = value;
            value = 0.0;
            for(int i=0; i<numPoints; ++i) {
                value += w[i]*function.getValue(start + x[i]*dx);
            }
            value *= dx;
            // If the solution is converged, return the
            // value. Otherwise, double the points in the
            // integration domain and try again.
            if ( Math.abs(value-oldValue) <= tolerance*Math.abs(oldValue) ) {
                return value;
            }
            numPoints *= 2;
            if (numPoints >= maxPoints) break;
        }
        // Solution did not converge
        return 12345.6;
    }

}

